from setuptools import setup

setup(
    name='kinepy',
    packages=['kinepy'],
    version='0.0.1',
    description='Python library dedicated to plane mechanism simulation',
    author='Loïc Chevalier - Valentin Burillier',
    url='https://github.com/valentin-burillier/kinepy',
    keywords=['robotics', 'dynamic', 'kinematics', 'static', 'inverse-kinematics', 'mechanism'],
    requires=[],
)
